/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * NodeAdapter.java
 *
 * Created on 10 marzo 2003, 12.32
 * An adapter for leaf nodes (Plugger would be a better name?)
 * @author  Valter Crescenzi
 */

package roadrunner.ast;

import java.util.List;
import java.util.ListIterator;
import java.util.Collections;

import roadrunner.util.Indenter;

public abstract class NodeAdapter implements Node, Cloneable {
  
  protected int id;
 
  protected NodeAdapter(int i) {
    this.id =  i;
  }
  
  protected int getTypeId() {
      return this.id;
  }
  
  public String getType() {
    return ASTConstants.jjtNodeName[this.id];
  }  
 
  public List jjtGetChildren() {
    return Collections.EMPTY_LIST;
  }
  
  public int jjtGetNumChildren() {
    return 0;
  }
  
  public Node jjtGetChild(int i) {
    throw new NoSuchNodeException("The child number "+i+" doesn\'t exist!");
  }
  
  /** Accept the visitor.  */
  public boolean jjtAccept(Visitor visitor) {
    throw new RuntimeException("This method should never be called!");
  }
  
  /** Accept the visitor. **/
  public boolean childrenAccept(Visitor visitor) {
    boolean result = true; //for result AND logic
    List children = jjtGetChildren();
    if (children != null) {
      ListIterator i = children.listIterator();
      while (i.hasNext()) {
        Node n = (Node)i.next();
        if (!n.jjtAccept(visitor)) result = false;
      }
    }
    return result;
  }
  
  public Node jjtGetNode(Node.Path path) {
    Node result = this;
    int indices[] = path.indices();
    for(int i=0; i<indices.length; i++) {
      result = result.jjtGetChild(indices[i]);
    }
    return result;
  }
  
  
  public Node clone(boolean deep) {
    NodeAdapter clone=null;
    try {
      clone = (NodeAdapter)clone();
    }
    catch (CloneNotSupportedException cnse) {
      System.out.print(cnse.getMessage());
      cnse.printStackTrace();
      System.exit(-1);
    }
    return clone;
  }
   
  /* Override this method if you want to customize how the node dumps out its children. */
  public String dump() {
    return dump(new Indenter(true));
  }
  public String dump(Indenter ind) {
    StringBuffer out = new StringBuffer();
    dump(out,ind);
    return out.toString();
  }
  
  private void dump(StringBuffer out, Indenter ind) {
    out.append(ind.toString()+this+"\n");
    ListIterator it = this.jjtGetChildren().listIterator();
    ind.inc();
    while (it.hasNext()) {
      NodeAdapter n = (NodeAdapter)it.next();
      n.dump(out, ind);
    }
    ind.dec();
  }
  
  public boolean equals(Object o) {
      NodeAdapter that = (NodeAdapter)o;
      return this.getTypeId() == that.getTypeId() && this.jjtGetChildren().equals(that.jjtGetChildren());
  }
  public int hashCode() {
      return this.getTypeId()+jjtGetChildren().hashCode();
  }
  
  public String toString() { return ASTConstants.jjtNodeName[id]; }
}
