/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * Lumberjack.java
 *
 * Created on 31 luglio 2003, 11.56
 */

package roadrunner.engine;

import java.util.*;
import java.util.logging.*;

import roadrunner.util.Util;
import roadrunner.ast.Expression;
import roadrunner.ast.ASTSubtree;
import roadrunner.ast.Node;
import roadrunner.ast.Walker;
import roadrunner.parser.TokenList;
import roadrunner.parser.Binder;
import roadrunner.parser.BindingException;

/**
 *
 * @author  Valter Crescenzi
 */
class Lumberjack {
    
    static private Logger log = Logger.getLogger(Lumberjack.class.getName());
    
    private Set expressions;
    private TokenList[] tokenlists;
    
    /** Creates a new instance of Lumberjack */
    Lumberjack(Set expressions, TokenList[] tl) {
        this.expressions = expressions;
        this.tokenlists = tl;
    }
    
    Set processSubtrees() {
        Set result = new LinkedHashSet();
        Iterator wIt = expressions.iterator();
        while (wIt.hasNext()) {
            try {
                Expression expression = (Expression)wIt.next();
                log.fine("Processing subtrees of expression:\n "+expression.dump("\t"));
                Expression processed = processSubtreesOf(expression);
                result.add(processed);
            }
            catch (BindingException be) {
                log.warning("This expression cannot processed: it does not match with its tokenlists");
            }
        }
        return result;
    }
    
    private Expression processSubtreesOf(Expression expression) throws BindingException {
        Binder binder = new Binder(expression);
        Map tokenlist2subtree = new LinkedHashMap(); // N.B. *must* preserve order
        Map subtree2path = getPathsOfSubtreeNodes(expression);
        SortedMap path2exp = new TreeMap();
        for (int i=0; i<tokenlists.length; i++) {
            binder.setBindings(tokenlists[i]);
            tokenlist2subtree.putAll(binder.getSubtreeBindings());
        }
        Map subtree2tokenlists = new IdentityHashMap();
        //N.B. if order is not fixed somehow, ordered of elements inside IdentityHashMap
        // depends on platform and even on runtime options of JVM
        Util.reverseMapping(subtree2tokenlists, tokenlist2subtree);
        Iterator subIt = subtree2tokenlists.entrySet().iterator();
        log.finer("Found "+subtree2tokenlists.size()+" subtrees to process");
        while (subIt.hasNext()) {
            Map.Entry entry = (Map.Entry)subIt.next();
            ASTSubtree subtree = (ASTSubtree) entry.getKey();
            List tokenlists = (List)entry.getValue();
            Node.Path path = (Node.Path)subtree2path.get(subtree);
            log.fine("Processing: "+subtree+" of path "+path);
            log.fine("Tokenlists associated with "+subtree+" of path "+path+"\n"+tokenlists.toString());
            Expression solution = matchSubtree(subtree, tokenlists);
            if (solution != null) {
                log.fine("Best replacing for subtree "+subtree+" of path "+path);
                log.fine("\n"+solution.dump(subtree.toString()+"\t"));
                path2exp.put(path,solution);
            }
            else {
                log.fine("No solutions: cannot remove "+subtree+" of path "+path);
            }
        }
        return expression.replaceSubtrees(path2exp);
    }
    
    private Expression matchSubtree(ASTSubtree subtree, List tls) {
        TokenList[] tlists = (TokenList[])tls.toArray(new TokenList[tls.size()]);
        Engine engine = new Engine(null, tlists, subtree.getRootToken().depth());
        Set solutions = engine.match();
        if (solutions.isEmpty()) return null;
        
        log.finer("Found "+solutions.size()+" replacing for "+subtree);
        log.finer("Saving the best and discarding all the others");
        Registry registry = new Registry(solutions, tlists);
        SortedSet orderedSolutions = registry.getExpressionsByCompressRatio();
        Expression best = (Expression)orderedSolutions.first();
        log.finer("Best solution CR: "+registry.getMDLEvaluator(best).getCompressRatio());
        return best;
    }
    
    private Map getPathsOfSubtreeNodes(final Expression expression) {
        final Map result = new IdentityHashMap();
        new Walker(expression) {
            public boolean visit(ASTSubtree node) {
                result.put(node, getCurrentAbsolutePath());
                return false;
            }
        }.visit(expression);
        return result;
    }
    
}
