/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * CharacteristicSet.java
 *
 * Created on 21 ottobre 2003, 12.01
 * @author  Valter Crescenzi
 */

package roadrunner.engine.sampler;

import java.util.Set;
import java.util.List;
import java.util.LinkedList;
import java.util.ListIterator;

import roadrunner.ast.Expression;
import roadrunner.bidi.Region;
import roadrunner.bidi.Direction;

public class CharacteristicSet {
    
    private Region region;
    private Direction dir;
    private ListIterator resultsIterator;
    private CharacteristicSampler sampler;
    private boolean alreadySampled; // All samples of the set have been already computed
    private int minDOMdepth;
    
    private List samples;
    
    /** Creates a new instance of CharacteristicSet */
    public CharacteristicSet(Expression exp) {
        this(exp.asRegion());
    }
    public CharacteristicSet(Region region) {
        this(region, Direction.LEFT2RIGHT);
    }
    public CharacteristicSet(Region region, Direction dir) {
        this(region,-1, dir);
    }
    public CharacteristicSet(Region region, int minDepth, Direction dir) {
        this.dir = dir;
        this.region = region;
        this.minDOMdepth = minDOMdepth;
        this.alreadySampled = false;
        this.sampler = new CharacteristicSampler(region, dir);
        this.sampler.setMinDOMdepth(minDepth);
        this.samples = new LinkedList();
        this.resultsIterator = null;
    }
    public CharacteristicSet(Region region, Set characteristicSet) {
        this.region = region;
        this.samples = new LinkedList(characteristicSet);
        this.alreadySampled = true;
    }
    
    public Region getRegion() {
        return this.region;
    }
    
    /**
     *  Lazy evaluation of characteristic samples
     **/
    public ListIterator listIterator() {
        if (this.alreadySampled) return this.samples.listIterator();
        
        this.resultsIterator = this.samples.listIterator();
        //next ListIterator implements lazy evaluation
        return new ListIterator() {
            public Object next() {
                if (!resultsIterator.hasNext())
                    computeNextIfAny();
                return resultsIterator.next();
            }
            public boolean hasNext() {
                if (!resultsIterator.hasNext())
                    computeNextIfAny();
                return resultsIterator.hasNext();
            }
            public int nextIndex() {
                return resultsIterator.nextIndex();
            }
            public Object previous() {
                return resultsIterator.previous();
            }
            public boolean hasPrevious() {
                return resultsIterator.hasPrevious();
            }
            public int previousIndex() {
                return resultsIterator.previousIndex();
            }
            public void add(Object o) { throw new UnsupportedOperationException(); }            
            public void remove()      { throw new UnsupportedOperationException(); }
            public void set(Object o) { throw new UnsupportedOperationException(); }
        };
    }
    
    private void computeNextIfAny() {
        CharacteristicSample sample = this.sampler.computeNext();
        if (sample!=null) {
            this.resultsIterator.add(sample);
            this.resultsIterator.previous(); // move the implicit cursor before the new element
        }
    }
    
    public CharacteristicSample first() {
        return (CharacteristicSample)listIterator().next();
    }
}
