/*
       Licensed to the Apache Software Foundation (ASF) under one
       or more contributor license agreements.  See the NOTICE file
       distributed with this work for additional information
       regarding copyright ownership.  The ASF licenses this file
       to you under the Apache License, Version 2.0 (the
       "License"); you may not use this file except in compliance
       with the License.  You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

       Unless required by applicable law or agreed to in writing,
       software distributed under the License is distributed on an
       "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
       KIND, either express or implied.  See the License for the
       specific language governing permissions and limitations
       under the License
*/
/*
 * TokenFactory.java
 *
 * Created on 5 febbraio 2003, 12.19
 * @author  Valter Crescenzi
 */

package roadrunner.parser.token;

import java.util.Set;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;


class TagFactory {
    // Flyweight Pattern for tag = element + attributes
    
    // Method equals() is used to compute a tag code which depends only on element and attribute *names*, not values
    // However, attribute values are kept since we need them to extract link and images. Therefore 2 tags are mapped
    // to the same object only if they have same, element, attribute names, and attribute values.
    // There may exist 2 different Tag object with the same tag code (it means that they differ for attribute values)
    
    private static TagFactory instance;
    
    static TagFactory getInstance() {
        if (instance==null) instance = new TagFactory();
        return instance;
    }
    
    private int lastcode;
    // first code2tag element is null because it corresponds to code 0 (PCDATA)
    private Map tag2code; // Map: Tag -> code
    
    private static Map element2attributes2tag; // Map:  (String ->  Map:  (Map -> Tag))
    
    private TagFactory() {
        this.lastcode = 0;
        this.tag2code = new HashMap();
        this.element2attributes2tag = new HashMap();
    }
    
    int getNumberOfKnownTags() {
        return lastcode;
    }
    
    Tag createTag(String el, Map attrs) {
        return getFlyweightTag(el, attrs);
    }
    
    private Tag getFlyweightTag(String el, Map attrs) {
        Map attributes2tag = (Map)element2attributes2tag.get(el);
        if (attributes2tag==null) return createNewFlyweightTagFromUnknownElement(el,attrs);
        Tag flyweight = (Tag)attributes2tag.get(attrs);
        if (flyweight==null) return createNewFlyweightTagFromUnknownAttributes(el,attrs);
        return flyweight; //Recycle objects
    }
    
    private Tag createNewFlyweightTagFromUnknownElement(String element, Map attributes)  {
        Integer code = new Integer(++lastcode);
        Tag result = new Tag(element, attributes);
        Map attributes2tag = new HashMap();
        attributes2tag.put(attributes, result);
        element2attributes2tag.put(element, attributes2tag);
        tag2code.put(result, code);
        return result;
    }
    
    private Tag createNewFlyweightTagFromUnknownAttributes(String element, Map attributes)  {
        Tag result = new Tag(element, attributes);
        Map attributes2tag = (Map)element2attributes2tag.get(element);
        attributes2tag.put(attributes, result);
        // Check whether a code for this tag already exists or not
        Integer known = (Integer)tag2code.get(result);
        if (known==null) { // need new code
            Integer code = new Integer(++lastcode);
            tag2code.put(result, code);
        }
        return result;
    }
    
    int getTagCode(Tag tag) {
        return ((Integer)tag2code.get(tag)).intValue();
    }
}
